'use strict';

class Game {
    constructor() {
        this._audio = new Audio();
        this._renderer = new Renderer();
        this._particleManager = new ParticleManager(this._renderer);
        this._input = new Input();

        this._audio.registerAudio('win', document.getElementById('audio_win'));
        this._audio.registerAudio('bomb_explodes', document.getElementById('audio_bomb_explodes'));
        this._audio.registerAudio('detection', document.getElementById('audio_detection'));
        this._audio.registerAudio('emit_search', document.getElementById('audio_emit_search'));
        this._audio.registerAudio('emit_sweep', document.getElementById('audio_emit_sweep'));
        this._audio.registerAudio('found_bomb', document.getElementById('audio_found_bomb'));
        this._audio.registerAudio('wasted_bombs', document.getElementById('audio_wasted_bombs'));

        this._renderer.registerSprite('checkmark', document.getElementById('sprite_checkmark'));
        this._renderer.registerSprite('bomb', document.getElementById('sprite_bomb'));
        this._renderer.registerSprite('explosion', document.getElementById('sprite_explosion'));
        this._renderer.registerSprite('flag', document.getElementById('sprite_flag'));
        this._renderer.registerSprite('cross', document.getElementById('sprite_cross'));
        this._renderer.registerSprite('sweep', document.getElementById('sprite_sweep'));
        this._renderer.registerSprite('search', document.getElementById('sprite_search'));
        this._renderer.registerSprite('mouse', document.getElementById('sprite_mouse'));

        this._levelCompleted = config.levels.map(() => false);

        this._levelIndex = 0;
        this._levelRunning = false;
        this._checkmarks = [];

        this._startLevel();
        this._startMainLoop();

        this._totalTime = 0.0;

        this._descriptionHidden = false;
        document.addEventListener('mousedown', (e) => {
            if (this._descriptionHidden) {
                return;
            }
            const desc = document.getElementById('description');
            desc.parentElement.removeChild(desc);
            this._descriptionHidden = true;
        });
    }

    _startMainLoop() {
        let last = null;
        function loop(time) {
            const elapsedSeconds = (last ? time - last : 16) / 1000.0;
            this._totalTime += elapsedSeconds;
            last = time;

            this._update(elapsedSeconds);

            requestAnimationFrame(loop.bind(this));
        }
        loop.bind(this)();
    }

    _startLevel() {
        this._level = config.levels[this._levelIndex];

        this._bombs = new Array(this._level.bombCount).fill(null).map(() => {
            return {
                position: Game._getRandomPosition(),
            };
        });
        this._crosses = [];
        this._flags = [];

        this._sweeps = [];
        this._detections = [];

        this._searchCount = this._level.searchCount;
        this._sweepCount = this._level.sweepCount;
        this._timerStartedAt = new Date();

        this._gameOverTimeout = setTimeout(() => {
            this._gameOver();
        }, this._level.timeLimitSeconds * 1000);

        this._randomSweepInterval = setInterval(() => {
            this._emitSweep(Game._getRandomPosition());
        }, this._level.randomSweepInterval * 1000);

        this._levelRunning = true;
    }

    _endLevel() {
        this._levelRunning = false;

        clearTimeout(this._gameOverTimeout);
        clearInterval(this._randomSweepInterval);
        setTimeout(() => {
            this._startLevel();
        }, config.cooldownSeconds * 1000);
    }

    _levelWon() {
        this._audio.playAudio('win');

        this._levelIndex += 1;

        if (this._levelIndex === config.levels.length) {
            alert('You unlucked the "I can not believe you just did that" achievement.');
            this._levelIndex = 0;
        }

        this._endLevel();

        this._checkmarks.push({
            position: [config.resolution.x / 2.0, config.resolution.y / 2.0],
            scale: 10.0,
        });
    }

    _gameOver() {
        this._audio.playAudio('bomb_explodes');

        this._bombs.forEach((bomb) => this._crosses.push({
            position: bomb.position,
        }));

        this._levelIndex = 0;
        this._checkmarks.length = 0;

        this._endLevel();

        for (let i = 0; i < 50; i++) {
            const scaleChange = Math.random() * 4.0;
            this._particleManager.createParticle('explosion', Math.random() * 3.0,
                Math.random() * config.resolution.x, Math.random() * config.resolution.y, 0.0, 0.0,
                0.0, 0.0, scaleChange, scaleChange, 0, -20 + Math.random() * 40);
        }
    }

    _emitSweep(position, isSearch=false) {
        this._sweeps.push({
            position: position,
            isSearch: isSearch,
            radius: 0,
            opacity: 1
        });
    }

    _update(elapsedSeconds) {
        this._renderer.clear();

        const fractionTimeOver = ((new Date() - this._timerStartedAt) / 1000.0) / this._level.timeLimitSeconds;

        this._renderer.drawBorder(this._totalTime * 20.0, fractionTimeOver > 0.5 ? '#ff0000' : '#00ff00', 6.0, 50.0);

        if (this._levelRunning) {
            this._renderer.drawSprite('mouse', config.resolution.x / 2.0, 40.0, 0.3, 0.3);

            for (let i = 0; i < this._sweepCount; i++) {
                this._renderer.drawSprite('sweep', config.resolution.x / 2.0 - 60.0 - i * 50.0, 40.0, 0.15, 0.15);
            }

            for (let i = 0; i < this._searchCount; i++) {
                this._renderer.drawSprite('search', config.resolution.x / 2.0 + 50.0 + i * 50.0, 40.0, 0.2, 0.2);
            }

            for (let i = 0; i < this._bombs.length; i++) {
                this._renderer.drawSprite('bomb', 30.0 + i * 40.0, config.resolution.y - 30.0, 0.15, 0.15);
            }

            const timeBarOffset = (fractionTimeOver * config.resolution.x) / 2.0;

            this._renderer.drawLineSegment(timeBarOffset, 72.0, config.resolution.x - timeBarOffset, 72.0, fractionTimeOver > 0.5 ? '#ff0000' : '#ffffff', 5.0);

            if (this._input.wasButtonPressed(0)) {
                if (this._sweepCount > 0) {
                    this._audio.playAudio('emit_sweep');

                    this._sweepCount--;
                    this._emitSweep(this._input.getWorldPosition(this._renderer));
                }
            }

            if (this._input.wasButtonPressed(2)) {
                if (this._searchCount > 0) {
                    this._audio.playAudio('emit_search');

                    this._searchCount--;
                    this._emitSweep(this._input.getWorldPosition(this._renderer), true);
                }
            }

            if (this._sweeps.length === 0 && this._searchCount === 0) {
                this._audio.playAudio('wasted_bombs');

                this._gameOver();
            } else if (this._bombs.length === 0) {
                this._levelWon();
            }
        }

        const destroyedBombs = new Set();

        this._sweeps.forEach((sweep) => {
            const oldRadius = sweep.radius;

            if (sweep.isSearch) {
                sweep.radius += config.searchExpandSpeed * elapsedSeconds;
                this._renderer.fillCircle(sweep.position[0], sweep.position[1], sweep.radius, '#8888ff');
                if (sweep.radius > config.searchFadeRadius) {
                    sweep.opacity -= config.searchFadeRate * elapsedSeconds;
                }
            } else {
                sweep.radius += config.sweepExpandSpeed * elapsedSeconds;
                this._renderer.drawCircle(sweep.position[0], sweep.position[1], sweep.radius, '#00ff00', 5 * sweep.opacity);
                if (sweep.radius > config.sweepFadeRadius) {
                    sweep.opacity -= config.sweepFadeRate * elapsedSeconds;
                }
            }

            this._bombs.forEach((bomb) => {
                const distance = Math.sqrt(
                    Math.pow(sweep.position[0] - bomb.position[0], 2) +
                    Math.pow(sweep.position[1] - bomb.position[1], 2));

                if (distance <= sweep.radius && distance > oldRadius) {
                    if (sweep.isSearch) {
                        this._audio.playAudio('found_bomb');

                        destroyedBombs.add(bomb);

                        this._flags.push({
                            position: bomb.position,
                        });
                    } else {
                        this._audio.playAudio('detection');

                        this._detections.push({
                            position: sweep.position,
                            radius: distance,
                            lifetime: config.detectionLifetime,
                        });
                    }
                }
            });
        });

        destroyedBombs.forEach((bomb) => this._bombs.splice(this._bombs.indexOf(bomb), 1));

        this._sweeps = this._sweeps.filter((sweep) => sweep.opacity > 0);

        this._detections.forEach((detection) => {
            this._renderer.drawCircle(detection.position[0], detection.position[1], detection.radius,
                '#ffff00', 5 * (detection.lifetime / config.detectionLifetime));
            detection.lifetime -= elapsedSeconds;
        });

        this._detections = this._detections.filter((detection) => detection.lifetime > 0);

        this._crosses.forEach((cross) => this._renderer.drawSprite('cross', cross.position[0], cross.position[1], 0.2, 0.2));
        this._flags.forEach((flag) => this._renderer.drawSprite('flag', flag.position[0], flag.position[1], 0.2, 0.2));
        this._checkmarks.forEach((checkmark, i) => {
            checkmark.scale = Game.lerp(checkmark.scale, 0.2, 0.05);
            checkmark.position[0] = Game.lerp(checkmark.position[0], config.resolution.x - 30 - i * 40, 0.02);
            checkmark.position[1] = Game.lerp(checkmark.position[1], config.resolution.y - 30, 0.02);

            this._renderer.drawSprite('checkmark', checkmark.position[0], checkmark.position[1], checkmark.scale, checkmark.scale)
        });

        this._particleManager.update(elapsedSeconds);
    }

    static _getRandomPosition() {
        return [config.horizontalMargin + Math.random() * (config.resolution.x - config.horizontalMargin * 2),
            config.topMargin + Math.random() * (config.resolution.y - (config.bottomMargin + config.topMargin))];
    }

    static lerp(v1, v2, t) {
        return v1 + (v2 - v1) * Math.max(0.0, Math.min(1.0, t));
    }
}
