'use strict';

class Renderer {
    constructor() {
        this._sprites = new Map();

        this._scale = 1.0;

        const canvas = document.getElementById('canvas');
        document.body.appendChild(canvas);

        canvas.oncontextmenu = (e) => e.preventDefault();

        this._context = canvas.getContext('2d');

        window.addEventListener('resize', this._handleCanvasResize.bind(this));

        this._handleCanvasResize();
    }

    clear() {
        this._context.clearRect(0, 0, this._context.canvas.width, this._context.canvas.height);
    }

    _handleCanvasResize() {
        const width = window.innerWidth;
        const height = window.innerHeight;

        this._context.canvas.width = config.resolution.x;
        this._context.canvas.height = config.resolution.y;

        const currentAspectRatio = width / height;
        const targetAspectRatio = config.resolution.x / config.resolution.y;

        if (Math.abs(currentAspectRatio - targetAspectRatio) < 0.000001) {
            return;
        }

        if (currentAspectRatio > targetAspectRatio) {
            // Horizontal letterboxes
            const lbWidth = (width - (height * targetAspectRatio)) / 2.0;
            this._scale = height / config.resolution.y;

            this._canvasOffset = [lbWidth, 0];

            this._context.canvas.style.width = `${config.resolution.x * this._scale}px`;
            this._context.canvas.style.height = `${height}px`;
            this._context.canvas.style.left = `${lbWidth}px`;
            this._context.canvas.style.top = '0px';
        } else {
            // Vertical letterboxes
            const lbHeight = (height - (width / targetAspectRatio)) / 2.0;
            this._scale = width / config.resolution.x;

            this._canvasOffset = [0, lbHeight];

            this._context.canvas.style.width = `${width}px`;
            this._context.canvas.style.height = `${config.resolution.y * this._scale}px`;
            this._context.canvas.style.left = '0px';
            this._context.canvas.style.top = `${lbHeight}px`;
        }
    }

    screenToWorld(screenPosition) {
        return [
            (screenPosition[0] - this._canvasOffset[0]) / this._scale,
            (screenPosition[1] - this._canvasOffset[1]) / this._scale];
    }

    registerSprite(spriteId, image) {
        this._sprites.set(spriteId, image);
    }

    getSpriteSize(spriteId) {
        const image = this._sprites.get(spriteId);
        return [image.naturalWidth, image.naturalHeight];
    }

    drawSprite(spriteId, x, y, scaleX = 1.0, scaleY = 1.0, rotation = 0.0) {
        const image = this._sprites.get(spriteId);

        this._context.save();

        this._context.translate(x, y);
        this._context.rotate(rotation);
        this._context.scale(scaleX, scaleY);

        this._context.drawImage(image, -image.naturalWidth / 2.0, -image.naturalHeight / 2.0);

        this._context.restore();
    }

    drawCircle(x, y, radius, strokeStyle, lineWidth) {
        this._context.save();

        this._context.strokeStyle = strokeStyle;
        this._context.lineWidth = lineWidth;

        this._context.beginPath();
        this._context.arc(x, y, radius, 0.0, 2.0 * Math.PI);
        this._context.stroke();

        this._context.restore();
    }

    fillCircle(x, y, radius, fillStyle) {
        this._context.save();

        this._context.fillStyle = fillStyle;

        this._context.beginPath();
        this._context.arc(x, y, radius, 0.0, 2.0 * Math.PI);
        this._context.fill();

        this._context.restore();
    }

    drawLineSegment(x0, y0, x1, y1, strokeStyle, lineWidth) {
        this._context.save();

        this._context.strokeStyle = strokeStyle;
        this._context.lineWidth = lineWidth;

        this._context.beginPath();
        this._context.moveTo(x0, y0);
        this._context.lineTo(x1, y1);
        this._context.stroke();

        this._context.restore();
    }

    drawBorder(offset, strokeStyle, lineWidth, segmentLength) {
        this._context.save();

        this._context.strokeStyle = strokeStyle;
        this._context.lineWidth = lineWidth;

        this._context.beginPath();

        if (this._canvasOffset[0] === 0) {
            for (let x = -segmentLength + offset % segmentLength; x < config.resolution.x; x += segmentLength) {
                this._context.moveTo(x, 0.0);
                this._context.lineTo(x + segmentLength / 2.0, 0.0);
                this._context.moveTo(config.resolution.x - x, config.resolution.y);
                this._context.lineTo(config.resolution.x - (x + segmentLength / 2.0), config.resolution.y);
            }
        } else {
            for (let y = -segmentLength + offset % segmentLength; y < config.resolution.y; y += segmentLength) {
                this._context.moveTo(0.0, y);
                this._context.lineTo(0.0, y + segmentLength / 2.0);
                this._context.moveTo(config.resolution.x, config.resolution.y - y);
                this._context.lineTo(config.resolution.x, config.resolution.y - (y + segmentLength / 2.0));
            }
        }

        this._context.stroke();

        this._context.restore();
    }
}
